import random, shelve, os
from decimal import Decimal

from geraldo.utils import get_attr_value, calculate_size, memoize
from geraldo.widgets import Widget, Label, SystemField
from geraldo.graphics import Graphic, RoundRect, Rect, Line, Circle, Arc,\
        Ellipse, Image
from geraldo.barcodes import BarCode
from geraldo.base import GeraldoObject, ManyElements
from geraldo.cache import CACHE_BY_QUERYSET, CACHE_BY_RENDER, CACHE_DISABLED,\
        make_hash_key, get_cache_backend
from geraldo.charts import BaseChart
from geraldo.exceptions import AbortEvent

class ReportPage(GeraldoObject):
    rect = None
    _elements = None
    width = None
    randomic_number = None

    def __init__(self):
        self._elements = []

        self.randomic_number = str(random.randint(1, 999999)).zfill(6)

    def get_children(self):
        return self._elements

    def add_element(self, el):
        """In future this method can be used to store pages on disk and reduce
        memory consuming"""
        self._elements.append(el)

    @property
    def elements(self):
        """In future this method can be used to restore pages from disk and reduce
        memory consuming"""
        for el in self._elements:
            yield el

    @memoize
    def repr_for_cache_hash_key(self):
        return '/'.join([el.repr_for_cache_hash_key() for el in self.elements
            if hasattr(el, 'repr_for_cache_hash_key')])

class ReportGenerator(GeraldoObject):
    """A report generator is used to generate a report to a specific format."""

    cache_enabled = None
    first_page_number = 1
    variables = None
    return_pages = False

    _is_first_page = True
    _is_latest_page = True
    _current_top_position = 0
    _current_left_position = 0
    _current_page_number = 1 # This variable is just used for generating, so, keep in mind it
                             # has't the current number while rendering
    _current_object = None
    _current_queryset = None
    _generation_datetime = None
    _highest_height = 0

    # Groupping
    _groups_values = None
    _groups_working_values = None
    _groups_changed = None
    _groups_stack = None

    # The rendered report has pages, each page is a ReportPage instance
    _rendered_pages = None
    _page_rect = None

    def __init__(self, report, first_page_number=1, variables=None, return_pages=False,
            pages=None, **kwargs):
        """This method should be overrided to receive others arguments"""
        self.report = report

        # Initializes some attributes
        self._rendered_pages = pages or []
        self._current_page_number = len(self._rendered_pages)
        self._groups_values = {}
        self._groups_working_values = {}
        self._groups_changed = {}
        self._groups_stack = []

        self.first_page_number = first_page_number
        self.variables = variables or self.variables or {}
        self.return_pages = return_pages

    def get_children(self):
        return self._rendered_pages

    def execute(self):
        """This method must be overrided to execute the report generation."""

        # Initializes pages
        self._is_first_page = True
 
    def render_border(self, borders_dict, rect_dict):
        """Renders a border in the coordinates setted in the rect."""
        b_all = borders_dict.get('all', None)
        if b_all:
            graphic = isinstance(b_all, Graphic) and b_all or Rect()
            graphic.set_rect(
                    left=rect_dict['left'],
                    top=rect_dict['top'] - rect_dict['height'],
                    width=rect_dict['right'] - rect_dict['left'],
                    height=rect_dict['height'],
                    )
            # If border is a number, it is recognized as the stroke width
            if isinstance(b_all, (int, float)):
                graphic.stroke_width = b_all

            self._rendered_pages[-1].add_element(graphic)

        b_left = borders_dict.get('left', None)
        if b_left:
            graphic = isinstance(b_left, Graphic) and b_left or Line()
            graphic.set_rect(
                    left=rect_dict['left'], top=rect_dict['top'],
                    right=rect_dict['left'], bottom=rect_dict['bottom']
                    )
            # If border is a number, it is recognized as the stroke width
            if isinstance(b_left, (int, float)):
                graphic.stroke_width = b_left

            self._rendered_pages[-1].add_element(graphic)

        b_top = borders_dict.get('top', None)
        if b_top:
            graphic = isinstance(b_top, Graphic) and b_top or Line()
            graphic.set_rect(
                    left=rect_dict['left'], top=rect_dict['top'],
                    right=rect_dict['right'], bottom=rect_dict['top']
                    )
            # If border is a number, it is recognized as the stroke width
            if isinstance(b_top, (int, float)):
                graphic.stroke_width = b_top

            self._rendered_pages[-1].add_element(graphic)

        b_right = borders_dict.get('right', None)
        if b_right:
            graphic = isinstance(b_right, Graphic) and b_right or Line()
            graphic.set_rect(
                    left=rect_dict['right'], top=rect_dict['top'],
                    right=rect_dict['right'], bottom=rect_dict['bottom']
                    )
            # If border is a number, it is recognized as the stroke width
            if isinstance(b_right, (int, float)):
                graphic.stroke_width = b_right

            self._rendered_pages[-1].add_element(graphic)

        b_bottom = borders_dict.get('bottom', None)
        if b_bottom:
            graphic = isinstance(b_right, Graphic) and b_right or Line()
            graphic.set_rect(
                    left=rect_dict['left'], top=rect_dict['bottom'],
                    right=rect_dict['right'], bottom=rect_dict['bottom']
                    )
            # If border is a number, it is recognized as the stroke width
            if isinstance(b_bottom, (int, float)):
                graphic.stroke_width = b_bottom

            self._rendered_pages[-1].add_element(graphic)

    def make_band_rect(self, band, top_position, left_position):
        """Returns the right band rect on the PDF canvas"""
        band_rect = {
                'left': left_position, #self.report.margin_left,
                'top': top_position,
                'right': left_position + self.calculate_size(band.width), #self.report.page_size[0] - self.report.margin_right,
                'bottom': top_position - self.calculate_size(band.height),
                'height': self.calculate_size(band.height),
                }
        return band_rect
 
    def make_widget_rect(self, widget, band_rect):
        """Returns the right widget rect on the PDF canvas"""
        widget_rect = {
                'left': band_rect['left'] + calculate_size(widget.left),
                'top': band_rect['top'] - calculate_size(widget.top),
                'right': band_rect['left'] + calculate_size(widget.left) + calculate_size(widget.width),
                'bottom': band_rect['top'] - calculate_size(widget.top) + calculate_size(widget.height),
                'height': calculate_size(widget.height),
                'width': calculate_size(widget.width),
                }
        return widget_rect

    def render_element(self, element, current_object, band, band_rect, temp_top,
            top_position):
        # Doesn't render not visible element
        if not element.visible:
            return

        # Widget element
        if isinstance(element, Widget):
            widget = element.clone()

            # Set widget colors
            widget.font_color = self.report.default_font_color

            # Set widget basic attributes
            widget.instance = current_object
            widget.generator = self
            widget.report = self.report # This should be done by a metaclass in Band domain TODO
            widget.band = band # This should be done by a metaclass in Band domain TODO
            widget.page = self._rendered_pages[-1]

            # Border rect
            widget_rect = self.make_widget_rect(widget, band_rect)

            if isinstance(widget, SystemField):
                widget.left = band_rect['left'] + self.calculate_size(widget.left)
                widget.top = self.calculate_top(temp_top, self.calculate_size(widget.top))

                temp_height = self.calculate_size(element.top) + self.calculate_size(widget.height)
            elif isinstance(widget, Label):
                para = self.make_paragraph(widget.text, self.make_paragraph_style(band, widget.style))

                if widget.truncate_overflow:
                    self.keep_in_frame(
                            widget,
                            self.calculate_size(widget.width),
                            self.calculate_size(widget.height),
                            [para],
                            mode='truncate',
                            )

                    widget.left = band_rect['left'] + self.calculate_size(widget.left)
                    widget.top = self.calculate_top(temp_top, self.calculate_size(widget.top), self.calculate_size(widget.height))
                else:
                    self.wrap_paragraph_on(para, self.calculate_size(widget.width), self.calculate_size(widget.height))
                    widget.left = band_rect['left'] + self.calculate_size(widget.left)
                    widget.top = self.calculate_top(temp_top, self.calculate_size(widget.top), self.calculate_size(para.height))

                temp_height = self.calculate_size(element.top) + self.calculate_size(para.height)
            else:
                temp_height = self.calculate_size(element.top) + self.calculate_size(widget.height)

            # Sets element height as the highest
            if temp_height > self._highest_height:
                self._highest_height = temp_height

            self._rendered_pages[-1].add_element(widget)

            # Borders
            self.render_border(widget.borders or {}, widget_rect)

        # Graphic element
        elif isinstance(element, Graphic):
            graphic = element.clone()

            # Set widget basic attributes
            graphic.instance = current_object
            graphic.generator = self
            graphic.report = self.report # This should be done by a metaclass in Band domain TODO
            graphic.band = band # This should be done by a metaclass in Band domain TODO
            graphic.page = self._rendered_pages[-1]

            # Set graphic colors
            graphic.fill_color = graphic.fill_color or self.report.default_fill_color
            graphic.stroke_color = graphic.stroke_color or self.report.default_stroke_color

            if isinstance(graphic, RoundRect):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top) - self.calculate_size(graphic.height)
            elif isinstance(graphic, Rect):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top) - self.calculate_size(graphic.height)
            elif isinstance(graphic, Line):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top)
                graphic.right = band_rect['left'] + self.calculate_size(graphic.right)
                graphic.bottom = top_position - self.calculate_size(graphic.bottom)
            elif isinstance(graphic, Circle):
                graphic.left_center = band_rect['left'] + self.calculate_size(graphic.left_center)
                graphic.top_center = top_position - self.calculate_size(graphic.top_center)
            elif isinstance(graphic, Arc):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top)
                graphic.right = band_rect['left'] + self.calculate_size(graphic.right)
                graphic.bottom = top_position - self.calculate_size(graphic.bottom)
            elif isinstance(graphic, Ellipse):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top)
                graphic.right = band_rect['left'] + self.calculate_size(graphic.right)
                graphic.bottom = top_position - self.calculate_size(graphic.bottom)
            elif isinstance(graphic, Image):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top) - self.calculate_size(graphic.height)
            elif isinstance(graphic, BarCode):
                barcode = graphic.render()
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top) - self.calculate_size(graphic.height)
                self.wrap_barcode_on(barcode, graphic.width, graphic.height)
            elif isinstance(element, BaseChart):
                graphic.left = band_rect['left'] + self.calculate_size(graphic.left)
                graphic.top = top_position - self.calculate_size(graphic.top) - self.calculate_size(graphic.height)

            # Sets element height as the highest
            temp_height = self.calculate_size(element.top) + self.calculate_size(graphic.height)
            if temp_height > self._highest_height:
                self._highest_height = temp_height

            self._rendered_pages[-1].add_element(graphic)

        # Many elements
        elif isinstance(element, ManyElements):
            # Set widget basic attributes
            element.instance = current_object
            element.generator = self
            element.report = self.report # This should be done by a metaclass in Band domain TODO
            element.band = band # This should be done by a metaclass in Band domain TODO
            element.page = self._rendered_pages[-1]

            # Get the elements and render them
            for el in element.get_elements():
                self.render_element(el, current_object, band, band_rect, temp_top, top_position)


    def render_band(self, band, top_position=None, left_position=None,
            update_top=True, current_object=None):
        """Generate a band having the current top position or informed as its
        top coordinate"""

        # Calls the before_print event
        try:
            band.do_before_print(generator=self)
        except AbortEvent:
            return False

        # Sets the current object
        current_object = current_object or self._current_object

        # Page width. This should be done in a metaclass in Report domain TODO
        self._rendered_pages[-1].width = self.calculate_size(self.report.page_size[0]) -\
                self.calculate_size(self.report.margin_left) - self.calculate_size(self.report.margin_right)

        # Default value for band width
        band.width = self.calculate_size(band.width) or self._rendered_pages[-1].width

        # Coordinates
        left_position = left_position or self.get_left_pos()

        # Increases the top position when being an inline displayed detail band
        if left_position > self.calculate_size(self.report.margin_left) and\
           getattr(band, 'display_inline', False) and\
           band.width < self.get_available_width():
            temp_height = band.height + getattr(band, 'margin_top', 0) + getattr(band, 'margin_bottom', 0)
            self.update_top_pos(decrease=self.calculate_size(temp_height))
        else:
            self.update_left_pos(set=0)
            left_position = self.get_left_pos()

        temp_top = top_position = top_position or self.get_top_pos()

        # Calculates the band dimensions on the canvas
        band_rect = self.make_band_rect(band, top_position, left_position)

        # Band borders
        self.render_border(band.borders, band_rect)

        # Variable that stores the highest height at all elements
        self._highest_height = 0

        # Loop at band widgets
        for element in band.elements:
            self.render_element(element, current_object, band, band_rect, temp_top,
                    top_position)

        # Updates top position
        if update_top:
            if band.auto_expand_height:
                band_height = self._highest_height
            else:
                band_height = self.calculate_size(band.height)

            band_height += self.calculate_size(getattr(band, 'margin_top', 0))
            band_height += self.calculate_size(getattr(band, 'margin_bottom', 0))

            self.update_top_pos(band_height)

        # Updates left position
        if getattr(band, 'display_inline', False):
            self.update_left_pos(band.width + self.calculate_size(getattr(band, 'margin_right', 0)))
        else:
            self.update_left_pos(set=0)

        # Child bands
        for child_band in band.child_bands or []: # TODO This "or []" here is a quickfix
            # Doesn't generate if it is not visible
            if not child_band.visible:
                continue

            self.in_tests = True # XXX
            self.force_blank_page_by_height(self.calculate_size(child_band.height))

            self.render_band(child_band)

        # Calls the before_print event
        band.do_after_print(generator=self)

        return True

    def force_blank_page_by_height(self, height):
        """Check if the height is in client available report height and
        makes a new page if necessary"""
        if Decimal(str(self.get_available_height())) < Decimal(str(height)):
            self.force_new_page()
            return True
        
        return False

    def append_new_page(self):
        self._rendered_pages.append(ReportPage())

    def force_new_page(self):
        """Starts a new blank page"""
        # Ends the current page
        self._current_top_position = 0

        # Starts a new one
        self.start_new_page()

        # Page footer
        self.render_page_footer()

    def render_begin(self):
        """Renders the report begin band if it exists"""
        if not self.report.band_begin:
            return

        # Doesn't generate this band if it is not visible
        if not self.report.band_begin.visible:
            return

        # Call method that print the band area and its widgets
        self.render_band(self.report.band_begin)

    def render_summary(self):
        """Generate the report summary band if it exists"""
        if not self.report.band_summary:
            return

        # Doesn't generate this band if it is not visible
        if not self.report.band_summary.visible:
            return

        # Clears groups stack
        self._groups_stack = []

        # Check to force new page if there is no available space
        self.force_blank_page_by_height(self.calculate_size(self.report.band_summary.height))

        # Call method that print the band area and its widgets
        self.render_band(self.report.band_summary)

    def render_page_header(self):
        """Generate the report page header band if it exists"""
        if not self.report.band_page_header:
            return

        # Doesn't generate this band if it is not visible
        if not self.report.band_page_header.visible:
            return

        # Call method that print the band area and its widgets
        self.render_band(
                self.report.band_page_header,
                top_position=self.calculate_size(self.report.margin_top),
                update_top=False,
                )

    def render_page_footer(self):
        """Generate the report page footer band if it exists"""
        if not self.report.band_page_footer:
            return

        # Doesn't generate this band if it is not visible
        if not self.report.band_page_footer.visible:
            return

        # Call method that print the band area and its widgets
        self.render_band(
                self.report.band_page_footer,
                top_position=self.calculate_size(self.report.page_size[1]) -\
                    self.calculate_size(self.report.margin_bottom) -\
                    self.calculate_size(self.report.band_page_footer.height),
                update_top=False,
                )

    def render_end_current_page(self):
        """Closes the current page, using page breaker constant. Everything done after
        this will draw into a new page. Before this, using the generate_page_footer
        method to draw the footer"""

        self.render_page_footer()

        if self._is_latest_page:
            self.render_summary()

        self._current_page_number += 1
        self._is_first_page = False
        self.update_top_pos(set=0) # <---- update top position
 
    def render_bands(self):
        """Loops into the objects list to create the report pages until the end"""
 
        # Preparing local auxiliar variables
        self._current_page_number = self.report.first_page_number
        self._current_object_index = 0
        objects = self.report.get_objects_list()

        # just an alias to make it shorter
        d_band = self.report.band_detail

        # Empty report
        if self.report.print_if_empty and not objects:
            self.start_new_page()
            self.render_begin()
            self.render_end_current_page()

        # Loop for pages
        while self._current_object_index < len(objects):
            # Starts a new page and generates the page header band
            self.start_new_page()
            first_object_on_page = True

            # Generate the report begin band
            if self._is_first_page:
                self.render_begin()

            # Does generate objects if there is no details band
            if not d_band:
                self._current_object_index = len(objects)

            # Loop for objects to go into grid on current page
            while self._current_object_index < len(objects):
                # Get current object from list
                self._current_object = objects[self._current_object_index]

                # Renders group bands for changed values
                self.calc_changed_groups(first_object_on_page)

                if not first_object_on_page:
                    # The current_object of the groups' footers is the previous 
                    # object, so we have access, in groups' footers, to the last
                    # object before the group breaking
                    self._current_object = objects[self._current_object_index-1]
                    self.render_groups_footers()
                    self._current_object = objects[self._current_object_index]

                self.render_groups_headers(first_object_on_page)

                # Generate this band only if it is visible
                # - "done True" means band was rendered ok
                # - "done False" means band rendering was aborted
                # - "done None" means band didn't render, but wasn't aborted
                if d_band.visible:
                    done = self.render_band(d_band)
                else:
                    done = None

                # Renders subreports
                if done != False:
                    self.render_subreports()

                # Next object
                self._current_object_index += 1
                first_object_on_page = False

                # Break this if this page doesn't suppport nothing more...
                # ... if there is no more available height
                if done != False:
                    if self.get_available_height() < self.calculate_size(d_band.height):
                        # right margin is not considered to calculate the necessary space
                        d_width = self.calculate_size(d_band.width) + self.calculate_size(getattr(d_band, 'margin_left', 0))

                        # ... and this is not an inline displayed detail band or there is no width available
                        if not getattr(d_band, 'display_inline', False) or self.get_available_width() < d_width:
                            break

                    # ... or this band forces a new page and this is not the last object in objects list
                    elif d_band.force_new_page and self._current_object_index < len(objects):
                        break

            # Sets this is the latest page or not
            self._is_latest_page = self._current_object_index >= len(objects)

            # Renders the finish group footer bands
            if self._is_latest_page:
                self.calc_changed_groups(False)
                self.render_groups_footers(force=True)

            # Ends the current page, printing footer and summary and necessary
            self.render_end_current_page()

            # Breaks if this is the latest item
            if self._is_latest_page:
                break

            # Increment page number
            self._current_page_number += 1

    def start_new_page(self, with_header=True):
        """Do everything necessary to be done to start a new page"""
        self.append_new_page()

        self.report.do_on_new_page(
                page=self._rendered_pages[-1],
                page_number=len(self._rendered_pages) + self.first_page_number - 1,
                generator=self,
                )

        if with_header:
            self.render_page_header()

        # Page borders
        if self.report.borders:
            if not self._page_rect:
                self._page_rect = self.report.get_page_rect()
                self._page_rect['top'] = self.calculate_size(self.report.page_size[1]) - self._page_rect['top']
                self._page_rect['bottom'] = self.calculate_size(self.report.page_size[1]) - self._page_rect['bottom']

            self.render_border(self.report.borders, self._page_rect)

    def calculate_size(self, size):
        """Uses the function 'calculate_size' to calculate a size"""
        return calculate_size(size)

    def get_left_pos(self):
        """Returns the left position of the drawer. Is useful on inline displayed detail bands"""
        return self.calculate_size(self.report.margin_left) + self._current_left_position

    def get_available_width(self):
        return self.calculate_size(self.report.page_size[0]) - self.calculate_size(self.report.margin_left) -\
                self.calculate_size(self.report.margin_right) - self._current_left_position

    def calculate_top(self, *args):
        return sum(args)

    def get_top_pos(self):
        """We use this to use this to get the current top position, 
        considering also the top margin."""
        ret = self.calculate_size(self.report.margin_top) + self._current_top_position

        if self.report.band_page_header:
            ret += self.calculate_size(self.report.band_page_header.height)

        return ret

    def get_available_height(self):
        """Returns the available client height area from the current top position
        until the end of page, considering the bottom margin."""
        ret = self.calculate_size(self.report.page_size[1]) - self.calculate_size(self.report.margin_bottom) -\
                self.calculate_size(self.report.margin_top) - self._current_top_position

        if self.report.band_page_header:
            ret -= self.calculate_size(self.report.band_page_header.height)

        if self.report.band_page_footer:
            ret -= self.calculate_size(self.report.band_page_footer.height)

        return ret

    def update_top_pos(self, increase=0, decrease=0, set=None):
        """Updates the current top position controller, increasing (by default),
        decreasing or setting it with a new value."""
        if set is not None:
            self._current_top_position = set
        else:        
            self._current_top_position += increase
            self._current_top_position -= decrease

        return self._current_top_position

    def update_left_pos(self, increase=0, decrease=0, set=None):
        """Updates the current left position controller, increasing (by default),
        decreasing or setting it with a new value."""
        if set is not None:
            self._current_left_position = set
        else:        
            self._current_left_position += increase
            self._current_left_position -= decrease

        return self._current_left_position

    def get_page_count(self):
        """Calculate and returns the page count for this report. The challenge
        here is do this calculate before to generate the pages."""
        return len(self._rendered_pages)

    def make_paragraph(self, text, style=None):
        """Uses the Paragraph class to return a new paragraph object"""
        raise Exception('Not implemented')

    def wrap_paragraph_on(self, paragraph, width, height):
        """Wraps the paragraph on the height/width informed"""
        raise Exception('Not implemented')

    def wrap_barcode_on(self, barcode, width, height):
        """Wraps the barcode on the height/width informed"""
        raise Exception('Not implemented')

    # Stylizing

    def set_fill_color(self, color):
        """Sets the current fill on canvas. Used for fonts and shape fills"""
        pass
    
    def set_stroke_color(self, color):
        """Sets the current stroke on canvas"""
        pass

    def set_stroke_width(self, width):
        """Sets the stroke/line width for shapes"""
        pass

    # Groups topic

    def calc_changed_groups(self, force_no_changed=False):
        """Defines which groups has been changed their driver values to be
        used to render group bands"""
        changed = force_no_changed

        # Stores the previous group values
        self._groups_working_values = self._groups_values.copy()

        # Loops on groups until find the first changed, then all under it are considered
        # changed also
        for group in self.report.groups:
            # Gets the current value to compare with the old one
            current_value = get_attr_value(self._current_object, group.attribute_name)

            # Set changed as True if if wasn't and there is a change
            changed = changed or current_value != self._groups_values.get(group, None)

            # Stores new values
            self._groups_changed[group] = changed
            self._groups_values[group] = current_value

            # Appends to the stack
            if changed:
                self._groups_stack.append(group)

    def render_groups_headers(self, first_object_on_page=False):
        """Renders the report headers using 'changed' definition calculated by
        'calc_changed_groups'"""

        # Update working values for groups
        self._groups_working_values = self._groups_values

        # Loops on groups to render changed ones
        new_page = False
        for group in self.report.groups:
            if self._groups_changed.get(group, None):
                # If there is no space for group header band, forces a new page
                if group.band_header and group.band_header.visible:
                    new_page = self.force_blank_page_by_height(self.calculate_size(group.band_header.height))

                # Forces a new page if this group is defined to do it
                if not new_page and group.force_new_page and self._current_object_index > 0 and not first_object_on_page:
                    self.render_page_footer()
                    self.force_new_page()

                # Renders the group header band
                if group.band_header and group.band_header.visible:
                    self.render_band(group.band_header)

    def render_groups_footers(self, force=False):
        """Renders the report footers using previous 'changed' definition calculated by
        'calc_changed_groups'"""

        # Loops on groups to render changed ones
        for group in reversed(self.report.groups):
            if force or ( self._groups_changed.get(group, None) and\
                          self._groups_stack and\
                          self._groups_stack[-1] == group ):
                if group.band_footer and group.band_footer.visible:
                    self.force_blank_page_by_height(self.calculate_size(group.band_footer.height))
                    self.render_band(group.band_footer)

                if self._groups_stack:
                    self._groups_working_values.pop(self._groups_stack[-1])

                self._groups_stack.pop()

    def get_current_queryset(self):
        """Returns the current queryset. This solves a problem with subreports
        footers and headers, and solves also flexibility and customization issues."""

        # Customized and SubReports
        if self._current_queryset is not None:
            return self._current_queryset

        # Groups
        elif self._groups_stack:
            return self.get_objects_in_group()

        # Defaul detail driver queryset
        return self.report.queryset

    def get_objects_in_group(self):
        """Returns objects filtered in the current group or all if there is no
        group"""

        filter_dict = dict([(group.attribute_name, value) for group, value in self._groups_working_values.items()])

        def filter_object(obj):
            for k,v in filter_dict.items():
                if get_attr_value(obj, k) != v:
                    return False

            return obj

        return filter(filter_object, self.report.queryset)

    # SubReports

    def render_subreports(self):
        """Renders subreports bands for the current object in, usings its
        own queryset.
        
        For a while just the detail band is rendered. Maybe in future we
        change this to accept header and footer."""

        def force_new_page(height):
            # Forces new page if there is no available space
            if self.get_available_height() < self.calculate_size(height):
                self.render_page_footer()
                self.force_new_page()

        for subreport in self.report.subreports:
            # Subreports must have detail band
            if not subreport.band_detail or not subreport.visible:
                continue

            # Sets the parent object and automatically clear the queryset
            # in memory
            subreport.parent_object = self._current_object

            # Sets the temporary currenty queryset
            self._current_queryset = subreport.get_objects_list()

            # Loops objects
            for num, obj in enumerate(self._current_queryset):
                # Renders the header band
                if num == 0 and subreport.band_header:
                    # Forces new page if there is no available space
                    force_new_page(subreport.band_header.height)

                    # Renders the header band
                    if subreport.band_header.visible:
                        self.render_band(subreport.band_header)

                # Forces new page if there is no available space
                force_new_page(subreport.band_detail.height)

                # Renders the detail band
                if subreport.band_detail.visible:
                    self.render_band(subreport.band_detail, current_object=obj)

            # Renders the footer band
            if subreport.band_footer:
                # Forces new page if there is no available space
                force_new_page(subreport.band_footer.height)

                # Renders the header band
                if subreport.band_footer.visible:
                    self.render_band(subreport.band_footer)

            # Sets back the default currenty queryset
            self._current_queryset = None

    def make_paragraph_style(self, band, style=None):
        """Merge report default_style + band default_style + widget style"""
        raise Exception('Not implemented')

    def keep_in_frame(self, widget, width, height, paragraphs, mode):
        raise Exception('Not implemented')

    def fetch_from_cache(self):
        if self.report.cache_status == CACHE_BY_QUERYSET:
            hash_key = self.get_hash_key(self.report.queryset)
        elif self.report.cache_status == CACHE_BY_RENDER:
            hash_key = self.get_hash_key(self._rendered_pages)

        cache = self.get_cache_backend()
        buffer = cache.get(hash_key)

        if buffer:
            # Write to file stream
            if hasattr(self.filename, 'write') and callable(self.filename.write):
                self.filename.write(buffer)
                return True
                
            # Write to file path
            elif isinstance(self.filename, basestring):
                fp = file(self.filename, 'w')
                fp.write(buffer)
                fp.close()
                return True

    def cached_before_render(self):
        """Check and loads the generated report from caching system before call method
        'render_bands'"""

        if not self.cache_enabled or self.report.cache_status != CACHE_BY_QUERYSET:
            return False

        return self.fetch_from_cache()

    def cached_before_generate(self):
        """Check and loads the generated report from caching system before call method
        'generate_pages'"""

        if not self.cache_enabled or self.report.cache_status != CACHE_BY_RENDER:
            return False

        return self.fetch_from_cache()

    def store_in_cache(self, content):
        """Sends the canvas content to write in the cache backend"""

        if not self.cache_enabled or self.report.cache_status == CACHE_DISABLED:
            return

        if self.report.cache_status == CACHE_BY_QUERYSET:
            hash_key = self.get_hash_key(self.report.queryset)
        elif self.report.cache_status == CACHE_BY_RENDER:
            hash_key = self.get_hash_key(self._rendered_pages)

        cache = self.get_cache_backend()

        return cache.set(hash_key, content)

    def get_hash_key(self, objects):
        """Calculates the hash_key, appending/prepending something if necessary"""
        return make_hash_key(self.report, objects)

    def get_cache_backend(self):
        return get_cache_backend(
                self.report.cache_backend,
                cache_file_root=self.report.cache_file_root,
                )




import datetime, types, decimal, re

try: 
    set 
except NameError: 
    from sets import Set as set     # Python 2.3 fallback 

from reportlab.lib.units import cm
from reportlab.lib.colors import black

from base import BAND_WIDTH, BAND_HEIGHT, Element, SubReport
from utils import get_attr_value, SYSTEM_FIELD_CHOICES, FIELD_ACTION_VALUE, FIELD_ACTION_COUNT,\
        FIELD_ACTION_AVG, FIELD_ACTION_MIN, FIELD_ACTION_MAX, FIELD_ACTION_SUM,\
        FIELD_ACTION_DISTINCT_COUNT
from exceptions import AttributeNotFound

class Widget(Element):
    """A widget is a value representation on the report"""
    _height = 0 #0.5*cm
    _width = 5*cm
    style = {}
    truncate_overflow = False
    
    get_value = None # A lambda function to get customized values

    instance = None
    report = None
    generator = None
    band = None
    borders = None

    def __init__(self, **kwargs):
        """This initializer is prepared to set arguments informed as attribute
        values."""
        for k,v in kwargs.items():
            setattr(self, k, v)

    def clone(self):
        new = super(Widget, self).clone()
        new.style = self.style
        new.truncate_overflow = self.truncate_overflow

        new.get_value = self.get_value

        new.instance = self.instance
        new.report = self.report
        new.generator = self.generator
        new.band = self.band
        new.borders = self.borders

        return new

class Label(Widget):
    """A label is just a simple text.
    
    'get_value' lambda must have 'text' argument."""
    _text = ''

    _repr_for_cache_attrs = ('text','left','top','height','width','style','visible')

    def _get_text(self):
        if self.get_value:
            try:
                return self.get_value(self, self._text)
            except TypeError:
                return self.get_value(self._text)

        return self._text

    def _set_text(self, value):
        self._text = value

    text = property(_get_text, _set_text)

    def clone(self):
        new = super(Label, self).clone()

        if not callable(self._text):
            new._text = self._text

        return new

EXP_QUOTED = re.compile('\w\(([^\'"].+?[^\'"])(|,.*?)\)')
EXP_QUOTED_SUB = re.compile('\(([^\'"].+?[^\'"])(|,.*?)\)')
EXP_TOKENS = re.compile('([\w\._]+|\*\*|\+|\-|\*|\/)')

class ObjectValue(Label):
    """This shows the value from a method, field or property from objects got
    from the queryset.
    
    You can inform an action to show the object value or an aggregation
    function on it.
    
    You can also use 'display_format' attribute to set a friendly string
    formating, with a mask or additional text.
    
    'get_value' and 'get_text' lambda attributes must have 'instance' argument.
    
    Set 'stores_text_in_cache' to False if you want this widget get its value
    and text on render and generate moments."""

    attribute_name = None
    action = FIELD_ACTION_VALUE
    display_format = '%s'
    objects = None
    get_text = None # A lambda function to get customized display values
    stores_text_in_cache = True
    expression = None
    converts_decimal_to_float = False
    converts_float_to_decimal = True
    _cached_text = None
    on_expression_error = None # Expected arguments:
                               #  - widget
                               #  - instance
                               #  - exception
                               #  - expression

    def __init__(self, *args, **kwargs):
        super(ObjectValue, self).__init__(*args, **kwargs)

        if self.expression:
            self.prepare_expression()

    def prepare_expression(self):
        if not self.expression:
            pass

        self.expression = self.expression.replace(' ','')

        while True:
            f = EXP_QUOTED.findall(self.expression)
            if not f:
                # Replace simple attribute name or method to value("")
                if '(' not in self.expression:
                    self.expression = 'value("%s")' % self.expression
                break

            self.expression = EXP_QUOTED_SUB.sub('("%s"%s)'%(f[0][0], f[0][1]), self.expression, 1)

    def get_object_value(self, instance=None, attribute_name=None):
        """Return the attribute value for just an object"""
        instance = instance or self.instance
        attribute_name = attribute_name or self.attribute_name

        # Checks lambda and instance
        if self.get_value and instance:
            try:
                return self.get_value(self, instance)
            except TypeError:
                return self.get_value(instance)

        # Checks this is an expression
        tokens = EXP_TOKENS.split(attribute_name)
        tokens = filter(bool, tokens) # Cleans empty parts
        if len(tokens) > 1:
            values = {}
            for token in tokens:
                if not token in ('+','-','*','/','**') and not  token.isdigit():
                    values[token] = self.get_object_value(instance, token)
            return eval(attribute_name, values)

        # Gets value with function
        value = get_attr_value(instance, attribute_name)

        # For method attributes --- FIXME: check what does this code here, because
        #                           get_attr_value has a code to do that, using
        #                           callable() checking
        if type(value) == types.MethodType:
            value = value()

        return value

    def get_queryset_values(self, attribute_name=None):
        """Uses the method 'get_object_value' to get the attribute value from
        all objects in the objects list, as a list"""

        objects = self.generator.get_current_queryset()
        return map(lambda obj: self.get_object_value(obj, attribute_name), objects)

    def _clean_empty_values(self, values):
        def clean(val):
            if not val:
                return 0
            elif isinstance(val, decimal.Decimal) and self.converts_decimal_to_float:
                return float(val)
            elif isinstance(val, float) and self.converts_float_to_decimal:
                return decimal.Decimal(str(val))
            
            return val

        return map(clean, values)

    def action_value(self, attribute_name=None):
        return self.get_object_value(attribute_name=attribute_name)

    def action_count(self, attribute_name=None):
        # Returns the total count of objects with valid values on informed attribute
        values = self.get_queryset_values(attribute_name)
        return len(filter(lambda v: v is not None, values))

    def action_avg(self, attribute_name=None):
        values = self.get_queryset_values(attribute_name)

        # Clear empty values
        values = self._clean_empty_values(values)

        return sum(values) / len(values)

    def action_min(self, attribute_name=None):
        values = self.get_queryset_values(attribute_name)
        return min(values)

    def action_max(self, attribute_name=None):
        values = self.get_queryset_values(attribute_name)
        return max(values)

    def action_sum(self, attribute_name=None):
        values = self.get_queryset_values(attribute_name)

        # Clear empty values
        values = self._clean_empty_values(values)

        return sum(values)

    def action_distinct_count(self, attribute_name=None):
        values = filter(lambda v: v is not None, self.get_queryset_values(attribute_name))
        return len(set(values))

    def action_coalesce(self, attribute_name=None, default=''):
        value = self.get_object_value(attribute_name=attribute_name)
        return value or unicode(default)

    def _text(self):
        if not self.stores_text_in_cache or self._cached_text is None:
            try: # Before all, tries to get the value using parent object
                value = self.band.get_object_value(obj=self)
            except AttributeNotFound:
                if self.expression:
                    value = self.get_value_by_expression()
                else:
                    value = getattr(self, 'action_'+self.action)()

            if self.get_text:
                try:
                    self._cached_text = unicode(self.get_text(self, self.instance, value))
                except TypeError:
                    self._cached_text = unicode(self.get_text(self.instance, value))
            else:
                self._cached_text = unicode(value)
            
        return self.display_format % self._cached_text

    def _set_text(self, value):
        self._cached_text = value

    text = property(lambda self: self._text(), _set_text)

    def clone(self):
        new = super(ObjectValue, self).clone()
        new.attribute_name = self.attribute_name
        new.action = self.action
        new.display_format = self.display_format
        new.objects = self.objects
        new.stores_text_in_cache = self.stores_text_in_cache
        new.expression = self.expression
        new.on_expression_error = self.on_expression_error

        return new

    def get_value_by_expression(self, expression=None):
        """Parses a given expression to get complex calculated values"""

        expression = expression or self.expression

        if not self.instance:
            global_vars = {}
        elif isinstance(self.instance, dict):
            global_vars = self.instance.copy()
        else:
            global_vars = self.instance.__dict__.copy()

        global_vars.update({
            'value': self.action_value,
            'count': self.action_count,
            'avg': self.action_avg,
            'min': self.action_min,
            'max': self.action_max,
            'sum': self.action_sum,
            'distinct_count': self.action_distinct_count,
            'coalesce': self.action_coalesce,
            })

        if isinstance(self.report, SubReport):
            global_vars.update({
                'parent': self.report.parent_object,
                'p': self.report.parent_object, # Just a short alias
                })

        try:
            return eval(expression, global_vars)
        except Exception, e:
            if not callable(self.on_expression_error):
                raise

            return self.on_expression_error(self, e, expression, self.instance)

class SystemField(Label):
    """This shows system informations, like the report title, current date/time,
    page number, pages count, etc.
    
    'get_value' lambda must have 'expression' and 'fields' argument."""
    expression = '%(report_title)s'

    fields = {
            'report_title': None,
            'page_number': None,
            'first_page_number': None,
            'last_page_number': None,
            'page_count': None,
            'current_datetime': None,
            'report_author': None,
        }

    def __init__(self, **kwargs):
        super(SystemField, self).__init__(**kwargs)

        # This is the safe way to use the predefined fields dictionary
        self.fields = SystemField.fields.copy()

        self.fields['current_datetime'] = datetime.datetime.now()

    def text(self):
        page_number = (self.fields.get('page_number') or self.generator._current_page_number) + self.generator.first_page_number - 1
        page_count = self.fields.get('page_count') or self.generator.get_page_count()

        fields = {
            'report_title': self.fields.get('report_title') or self.report.title,
            'page_number': page_number,
            'first_page_number': self.generator.first_page_number,
            'last_page_number': page_count + self.generator.first_page_number - 1,
            'page_count': page_count,
            'current_datetime': self.fields.get('current_datetime') or datetime.datetime.now(),
            'report_author': self.fields.get('report_author') or self.report.author,
        }
        
        if self.get_value:
            return self.get_value(self.expression, fields)

        return self.expression%SystemFieldDict(self, fields)
    text = property(text)

    def clone(self):
        new = super(SystemField, self).clone()
        new.expression = self.expression
        new.fields = self.fields

        return new

class SystemFieldDict(dict):
    widget = None
    fields = None

    def __init__(self, widget, fields):
        self.widget = widget
        self.fields = fields or {}

        super(SystemFieldDict, self).__init__(**fields)

    def __getitem__(self, key):
        if key.startswith('now:'):
            return self.widget.report.format_date(
                    self.fields.get('current_datetime', datetime.datetime.now()),
                    key[4:]
                    )

        elif key.startswith('var:'):
            return self.widget.report.get_variable_value(name=key[4:], system_fields=self)

        return self.fields[key]




import datetime, csv
from base import ReportGenerator

from geraldo.utils import get_attr_value, calculate_size
from geraldo.widgets import Widget, Label, SystemField, ObjectValue
from geraldo.graphics import Graphic, RoundRect, Rect, Line, Circle, Arc,\
        Ellipse, Image
from geraldo.exceptions import AbortEvent

class CSVGenerator(ReportGenerator):
    """This is a generator to output data in CSV format. This format can be imported as a
    spreadsheet to Excel, OpenOffice Calc, Google Docs Spreadsheet, and others.
    
    Attributes:

        * 'filename' - is the file path you can inform optionally to save text to.
        * 'writer' - is csv.writer function you can inform manually to make it customizable.
                     This function must expects a first argument to receive a file object and
                     returns a csv.writer object.
    """
    writer = None
    writer_function = csv.writer
    first_row_with_column_names = False

    mimetype = 'text/csv'

    def __init__(self, report, cache_enabled=None, writer=None, first_row_with_column_names=None, **kwargs):
        super(CSVGenerator, self).__init__(report, **kwargs)

        # Cache enabled
        if cache_enabled is not None:
            self.cache_enabled = cache_enabled
        elif self.cache_enabled is None:
            self.cache_enabled = bool(self.report.cache_status)

        # Sets the writer function
        self.writer = writer or self.writer

        # Sets to append the first row with column names (ObjectValue name/attribute_name/expression)
        if first_row_with_column_names is not None:
            self.first_row_with_column_names = first_row_with_column_names

        # Additional attributes
        for k,v in kwargs.items():
            setattr(self, k, v)

    def start_writer(self, filename=None):
        if self.writer:
            return

        filename = filename or self.filename

        if isinstance(filename, basestring):
            filename = file(filename, 'w')

        # Default writer uses comma as separator and quotes only when necessary
        self.writer = self.writer_function(filename, quoting=csv.QUOTE_MINIMAL)

    def execute(self):
        super(CSVGenerator, self).execute()

        # Calls the before_print event
        self.report.do_before_print(generator=self)

        # Write the CSV output
        self.generate_csv()

        # Calls the after_print event
        self.report.do_after_print(generator=self)

    def get_hash_key(self, objects):
        """Appends pdf extension to the hash_key"""
        return super(CSVGenerator, self).get_hash_key(objects) + '.csv'

    # METHODS THAT ARE TOTALLY SPECIFIC TO THIS GENERATOR AND MUST
    # OVERRIDE THE SUPERCLASS EQUIVALENT ONES

    def generate_csv(self):
        """Generates the CSV output"""

        self._current_object_index = 0
        objects = self.report.get_objects_list()

        self.start_writer()

        # Make a sorted list of columns
        columns = [el for el in self.report.band_detail.elements if isinstance(el, ObjectValue)]
        columns.sort(lambda a,b: cmp(a.left, b.left) or cmp(a.width, b.width))

        # First row with column names
        if self.first_row_with_column_names:
            cells = [(col.name or col.expression or col.attribute_name) for col in columns]
            self.writer.writerow(cells)

        while self._current_object_index < len(objects):
            # Get current object from list
            self._current_object = objects[self._current_object_index]

            cells = []

            for element in columns:
                widget = element.clone()

                # Set widget colors
                widget.font_color = self.report.default_font_color

                # Set widget basic attributes
                widget.instance = self._current_object
                widget.generator = self
                widget.report = self.report
                widget.band = self.report.band_detail
                widget.page = None

                cells.append(widget.text)

            # Next object
            self._current_object_index += 1

            self.writer.writerow(cells)




import re, random, decimal

from reportlab.graphics.shapes import Drawing, String
from reportlab.graphics.charts.barcharts import HorizontalBarChart as OriginalHorizBarChart
from reportlab.graphics.charts.barcharts import VerticalBarChart as OriginalVertBarChart
from reportlab.graphics.charts.barcharts import HorizontalBarChart3D as OriginalHorizBarChart3D
from reportlab.graphics.charts.barcharts import VerticalBarChart3D as OriginalVertBarChart3D
from reportlab.graphics.charts.doughnut import Doughnut as OriginalDoughnutChart
from reportlab.graphics.charts.linecharts import HorizontalLineChart as OriginalLineChart
from reportlab.graphics.charts.piecharts import Pie as OriginalPieChart
from reportlab.graphics.charts.spider import SpiderChart as OriginalSpiderChart
from reportlab.graphics.charts.legends import Legend
from reportlab.lib.colors import HexColor, getAllNamedColors

from utils import cm, memoize, get_attr_value
from cross_reference import CrossReferenceMatrix, CROSS_COLS, CROSS_ROWS
from graphics import Graphic

DEFAULT_TITLE_HEIGHT = 1*cm

class BaseChart(Graphic):
    """Abstract chart class"""

    chart_class = None
    title = None
    colors = None
    _width = 8*cm
    _height = 7*cm
    rows_attribute = None
    cols_attribute = None
    cell_attribute = None
    action = 'first'
    data = None
    chart_style = None # Additional chart attributes
    axis_labels = None
    axis_labels_angle = None
    legend_labels = False
    values_labels = ' %s '
    replace_none_by_zero = True
    round_values = False
    summarize_by = None # Can be None, CROSS_ROWS or CROSS_COLS

    def __init__(self, **kwargs):
        # Set instance attributes
        for k,v in kwargs.items():
            if k == 'style':
                setattr(self, 'chart_style', v)
            else:
                setattr(self, k, v)

        # Prepare the title
        if self.title:
            self.title = isinstance(self.title, dict) and self.title or {'text': self.title}
            self.title.setdefault('fontSize', 14)
            self.title.setdefault('textAnchor', 'middle')
            self.title.setdefault('height', DEFAULT_TITLE_HEIGHT)

        # Prepare the colors
        if self.colors == False:
            self.legend_labels = None
        elif not self.colors:
            self.colors = self.get_available_colors()
        else:
            self.prepare_colors()

        # Prepare chart additional kwargs
        self.chart_style = self.chart_style or {}

    def clone(self):
        new = super(BaseChart, self).clone()

        new.chart_class = self.chart_class
        new.title = self.title
        new.colors = self.colors
        new.rows_attribute = self.rows_attribute
        new.cols_attribute = self.cols_attribute
        new.cell_attribute = self.cell_attribute
        new.action = self.action
        new.data = self.data
        new.chart_style = self.chart_style
        new.axis_labels = self.axis_labels
        new.axis_labels_angle = self.axis_labels_angle
        new.legend_labels = self.legend_labels
        new.values_labels = self.values_labels
        new.replace_none_by_zero = self.replace_none_by_zero
        new.round_values = self.round_values
        new.summarize_by = self.summarize_by

        return new

    # DRAWING METHODS

    @memoize
    def get_available_colors(self):
        """Returns a list of available colors"""

        # Get reportlab available colors
        colors = getAllNamedColors()

        # Remove bad colors
        colors.pop('white', None)
        colors.pop('black', None)

        # Returns only the colors values (without their names)
        colors = colors.values()
        
        # Shuffle colors list
        random.shuffle(colors)

        return colors

    def prepare_colors(self):
        colors = []

        for color in self.colors:
            try:
                colors.append(HexColor(color))
            except ValueError:
                pass

        self.colors = colors + self.get_available_colors()

    def get_drawing(self, chart):
        """Create and returns the drawing, to be generated"""

        drawing = Drawing(self.width, self.height)

        # Make the title
        title = self.make_title(drawing)

        # Setting chart dimensions
        chart.height = self.height
        chart.width = self.width

        # Make the legend
        legend = self.make_legend(drawing, chart)

        if title:
            chart.height -= self.title.get('height', DEFAULT_TITLE_HEIGHT)
            self.top += self.title.get('height', DEFAULT_TITLE_HEIGHT)

        # Setting additional chart attributes
        self.set_chart_style(chart)

        # Adds the chart to drawing to return
        drawing.add(chart)

        # Resizes to make sure everything is fitting
        drawing = drawing.resized()

        return drawing

    def make_legend(self, drawing, chart):
        if not self.legend_labels:
            return

        # Get legend labels
        labels = self.get_legend_labels()

        # Legend object
        legend = Legend()

        legend.colorNamePairs = zip(self.colors[:len(labels)], labels)
        legend.columnMaximum = len(legend.colorNamePairs)
        legend.deltay = 5
        legend.alignment = 'right'
        legend.x = drawing.width + 40
        legend.y = drawing.height - (self.title and self.title.get('height', DEFAULT_TITLE_HEIGHT) or 0)

        # Sets legend extra attributes if legend_labels is a dictionary
        if isinstance(self.legend_labels, dict):
            for k,v in self.legend_labels.items():
                if k != 'labels' and v:
                    setattr(legend, k, v)

        drawing.add(legend)

        return legend

    def get_legend_labels(self):
        # Use same axis if is summarizing
        if self.summarize_by:
            return self.get_axis_labels()

        # Base labels
        if isinstance(self.legend_labels, dict) and self.legend_labels.get('labels', None):
            labels = self.legend_labels['labels']
        elif isinstance(self.legend_labels, (tuple,list)):
            labels = self.legend_labels
        else:
            labels = self.get_cross_data().rows()

        # Calculated labels
        if callable(self.legend_labels):
            labels = [self.legend_labels(self, label, num) for num, label in enumerate(labels)]
        elif isinstance(self.legend_labels, basestring):
            labels = [self.get_cross_data().first(self.legend_labels, col=label) for label in labels]

        return map(unicode, labels)

    def get_axis_labels(self):
        # Base labels
        if isinstance(self.axis_labels, dict) and self.axis_labels.get('labels', None):
            labels = self.axis_labels['labels']
        elif isinstance(self.axis_labels, (tuple,list)):
            labels = self.axis_labels
        elif self.summarize_by == CROSS_ROWS:
            labels = self.get_cross_data().rows()
        else:
            labels = self.get_cross_data().cols()

        # Calculated labels
        if callable(self.axis_labels):
            labels = [self.axis_labels(self, label, num) for num, label in enumerate(labels)]
        elif isinstance(self.axis_labels, basestring):
            if self.summarize_by == CROSS_ROWS:
                labels = [self.get_cross_data().first(self.axis_labels, row=label) for label in labels]
            else:
                labels = [self.get_cross_data().first(self.axis_labels, col=label) for label in labels]

        return map(unicode, labels)

    def make_title(self, drawing):
        if not self.title:
            return

        # Make the dict with kwargs
        kwargs = self.title.copy()
        kwargs.setdefault('x', drawing.width / 2)
        kwargs.setdefault('y', drawing.height)

        # Make the string
        title = String(**kwargs)

        drawing.add(title)

        return title

    # CHART METHODS

    def get_cross_data(self, data=None):
        if not getattr(self, '_cross_data', None):
            data = data or self.data

            # Transforms data to cross-reference matrix
            if isinstance(data, basestring):
                data = get_attr_value(self.instance, data)

            if not isinstance(data, CrossReferenceMatrix):
                if self.rows_attribute: # and self.cols_attribute:
                    data = CrossReferenceMatrix(
                            data,
                            self.rows_attribute,
                            self.cols_attribute,
                            decimal_as_float=True,
                            )

            self._cross_data = data

        return self._cross_data

    def get_data(self):
        data = self.data

        # Returns nothing data is empty
        if not data:
            data = self.report.queryset # TODO: Change to support current objects
                                        # list (for subreports and groups)

        # Transforms data to cross-reference matrix
        data = self.get_cross_data(data)

        # Summarize data or get its matrix (after it is a Cross-Reference Matrix)
        if self.summarize_by == CROSS_ROWS:
            data = data.summarize_rows(self.cell_attribute, self.action)
        elif self.summarize_by == CROSS_COLS:
            data = data.summarize_cols(self.cell_attribute, self.action)
        else:
            data = data.matrix(self.cell_attribute, self.action)

        def none_to_zero(value):
            if value is None:
                value = 0
            elif isinstance(value, (list, tuple)):
                value = [cell or 0 for cell in value]

            return value

        def round_values(value):
            if isinstance(value, (float, decimal.Decimal)):
                value = int(round(value))
            elif isinstance(value, (list, tuple)):
                value = map(int, map(round, value))

            return value

        # Replace None to Zero
        if self.replace_none_by_zero:
            data = map(none_to_zero,  data)

        # Truncate decimal places
        if self.round_values:
            data = map(round_values,  data)

            # Stores major value in temporary variable to use it later
            if data:
                if isinstance(data[0], int):
                    self._max_value = max(data)
                elif isinstance(data[0], (list, tuple)):
                    self._max_value = max(map(max, data))

        return data

    def set_chart_attributes(self, chart):
        # Cols (Y) labels - Y axis
        if self.axis_labels:
            chart.categoryAxis.categoryNames = self.get_axis_labels()

            if self.axis_labels_angle is not None:
                chart.categoryAxis.labels.angle = self.axis_labels_angle
                chart.categoryAxis.labels.boxAnchor = 'ne'

    def set_chart_style(self, chart):
        # Setting additional chart attributes
        if self.chart_style:
            for k,v in self.chart_style.items():
                setattr(chart, k, v)

    def create_chart(self):
        chart = self.chart_class()

        return chart

    def render(self):
        # Make data matrix
        data = self.get_data()

        if not data:
            return

        # Creates the chart instance
        chart = self.create_chart()
        chart.data = data

        # Sets additional attributes
        self.set_chart_attributes(chart)

        return self.get_drawing(chart)

class BaseMatrixChart(BaseChart):
    """Abstract chart class to support matrix charts"""

    def get_data(self):
        data = super(BaseMatrixChart, self).get_data()

        if data and self.summarize_by:
            data = [data]
        
        return data

class LineChart(BaseMatrixChart):
    chart_class = OriginalLineChart

    def set_chart_attributes(self, chart):
        super(LineChart, self).set_chart_attributes(chart)

        # Cells labels
        if isinstance(self.values_labels, (tuple, list)):
            self.chart_style.setdefault('lineLabelFormat', self.values_labels)
        elif isinstance(self.values_labels, dict) and self.values_labels.get('labels', None):
            self.chart_style.setdefault('lineLabelFormat', self.values_labels['labels'])
        else:
            self.chart_style.pop('lineLabelFormat', None)

        # Set the line colors
        if self.colors:
            for num, color in enumerate(self.colors):
                try:
                    chart.lines[num].strokeColor = color
                except IndexError:
                    break

        # Value Axis min value
        if getattr(self, 'y_axis_min_value', None) != None:
            chart.valueAxis.valueMin = self.y_axis_min_value

        # Informed value axis step value
        if getattr(self, 'y_axis_step_value', None):
            chart.valueAxis.valueStep = self.y_axis_step_value

        # Value axis without decimal values
        elif self.round_values and getattr(self, '_max_value', None):
            chart.valueAxis.valueStep = round(self._max_value / 4)

class BarChart(BaseMatrixChart):
    chart_class = None
    horizontal = False # If is not horizontal, is because it is vertical (default)
    is3d = False

    def __init__(self, *args, **kwargs):
        super(BarChart, self).__init__(*args, **kwargs)

        # Chart class varies depending on attributes
        if not self.chart_class:
            if self.horizontal and self.is3d:
                self.chart_class = OriginalHorizBarChart3D
            elif self.horizontal:
                self.chart_class = OriginalHorizBarChart
            elif self.is3d:
                self.chart_class = OriginalVertBarChart3D
            else:
                self.chart_class = OriginalVertBarChart

    def clone(self):
        new = super(BarChart, self).clone()

        new.horizontal = self.horizontal
        new.is3d = self.is3d

        return new

    def set_chart_attributes(self, chart):
        super(BarChart, self).set_chart_attributes(chart)

        # Cells labels
        if self.values_labels:
            if isinstance(self.values_labels, (tuple, list)):
                self.chart_style.setdefault('barLabelFormat', self.values_labels)
            elif isinstance(self.values_labels, dict) and self.values_labels.get('labels', None):
                self.chart_style.setdefault('barLabelFormat', self.values_labels['labels'])

            # Label orientation
            if self.horizontal:
                chart.barLabels.boxAnchor = 'w'
            else:
                chart.barLabels.boxAnchor = 's'
        else:
            self.chart_style.pop('barLabelFormat', None)

        # Set bar strokes
        chart.bars.strokeWidth = 0

        # Forces bars to start from 0 (instead of lower value)
        chart.valueAxis.forceZero = 1

        # Shows axis X labels
        if not self.summarize_by: # XXX
            chart.categoryAxis.categoryNames = self.get_axis_labels()

        # Set the bar colors
        if self.colors:
            for num, color in enumerate(self.colors):
                try:
                    chart.bars[num].fillColor = color
                except IndexError:
                    break

    def get_data(self):
        data = super(BarChart, self).get_data()

        # Forces multiple colors
        if self.summarize_by and data:
            data = [[i] for i in data[0]]

        return data

class HorizontalBarChart(BarChart):
    horizontal = True

class SpiderChart(BaseMatrixChart):
    chart_class = OriginalSpiderChart

    def set_chart_attributes(self, chart):
        # Chart labels
        chart.labels = self.get_axis_labels()

        # Set the strands colors
        if self.colors:
            for num, color in enumerate(self.colors):
                try:
                    chart.strands[num].fillColor = color
                except IndexError:
                    break

class PieChart(BaseChart):
    chart_class = OriginalPieChart
    slice_popout = None

    def __init__(self, **kwargs):
        super(PieChart, self).__init__(**kwargs)

        # Force default value for summarize
        if not self.summarize_by:
            self.summarize_by = CROSS_ROWS

    def set_chart_attributes(self, chart):
        # Sets the slice colors
        if self.colors:
            for num, color in enumerate(self.colors):
                try:
                    chart.slices[num].fillColor = color
                except IndexError:
                    break

        # Sets the slice to popout
        pos = -1
        if self.slice_popout == True:
            data = self.get_data()
            pos = data.index(max(data))
        elif isinstance(self.slice_popout, int):
            pos = self.slice_popout
        elif callable(self.slice_popout):
            pos = self.slice_popout(self, chart)

        if pos >= 0:
            chart.slices[pos].popout = 20

        # Default labels
        chart.labels = self.get_axis_labels()

        # Cells labels
        if isinstance(self.values_labels, dict):
            for k,v in self.values_labels.items():
                if k == 'labels' and v:
                    chart.labels = v
                else:
                    setattr(chart.slices, k, v)

    def clone(self):
        new = super(PieChart, self).clone()
        new.slice_popout = self.slice_popout
        return new

    def get_drawing(self, chart):
        if self.action == 'percent':
            chart.labels = ['%s - %s%%'%(label,val) for label, val in zip(chart.labels, chart.data)]

        return super(PieChart, self).get_drawing(chart)

class DoughnutChart(PieChart):
    chart_class = OriginalDoughnutChart

